/* $Header: /belch_a/users/rearl/tinymuck/include/RCS/externs.h,v 1.12 90/09/18 08:06:03 rearl Exp $
 * 
 * $Log:	externs.h,v $
 * Revision 1.12  90/09/18  08:06:03  rearl
 * Added hash tables for players and primitives.
 * 
 * Revision 1.11  90/09/13  06:33:47  rearl
 * do_dump() and do_toad() prototypes changed.
 * 
 * Revision 1.10  90/09/10  02:24:17  rearl
 * Changed NL line termination to CR/LF pairs.
 * 
 * Revision 1.9  90/09/01  06:03:56  rearl
 * Put do_drop() back like it was.
 * 
 * Revision 1.8  90/08/27  14:06:53  rearl
 * Added support for environments: @trace, mods to drop and @dig commands.
 * 
 * Revision 1.7  90/08/15  03:11:00  rearl
 * Messed around with pronoun_substitute, just like in stringutil.c.
 * 
 * Revision 1.6  90/08/06  02:47:24  rearl
 * Put can_link() back, added restricted() from predicates.c.
 * 
 * Revision 1.5  90/07/30  00:10:47  rearl
 * Added @owned command.
 * 
 * Revision 1.4  90/07/29  17:18:22  rearl
 * Fixed the functions containing the = bug.
 * 
 * Revision 1.3  90/07/20  11:43:06  casie
 * Added new function declarations for log.c
 * 
 * Revision 1.2  90/07/19  23:14:18  casie
 * Removed log comments from top.
 * 
 */
#include "copyright.h"

#include "db.h"

/* Prototypes for externs not defined elsewhere */

extern char match_args[];

/* from db.c */
extern struct plist *new_prop(void);
extern void free_prop(struct plist *p);
extern int  number(const char *s);
extern void putproperties(FILE *f, struct plist *p);
extern void putprop(FILE *f, struct plist *p);
extern struct plist *getproperties(FILE *f);
extern void free_line(struct line *l);
extern void db_free_object(dbref i);
extern void db_clear_object(dbref i);
extern void macrodump(struct macrotable *node, FILE *f);
extern void macroload();
extern void db_grow(dbref newtop);

/* From create.c */
extern void do_open(dbref player, const char *direction, const char *linkto);
extern void do_link(dbref player, const char *name, const char *room_name);
extern void do_dig(dbref player, const char *name, const char *pname);
extern void do_create(dbref player, char *name, int cost);
extern void do_prog(dbref player, const char *name);
extern int unset_source(dbref player, dbref loc, dbref action);
/* from edit.c */
struct macrotable
    *new_macro(const char *name, const char *definition, dbref player);
extern void match_and_list(dbref player, const char *name, char *linespec);

/* From game.c */
extern void do_dump(dbref player, const char *newfile);
extern void do_shutdown(dbref player);

/* From hashtab.c */
extern hash_data *find_hash(const char *s, hash_tab *table, unsigned size);
extern hash_entry *add_hash(const char *name, hash_data data, hash_tab *table,
                            unsigned size);
extern int free_hash(const char *name, hash_tab *table, unsigned size);
extern void kill_hash(hash_tab *table, unsigned size);

/* From help.c */
extern void spit_file(dbref player, const char *filename);
extern void do_help(dbref player);
extern void do_news(dbref player);
extern void do_help(dbref player);

/* From look.c */
extern void look_room(dbref player, dbref room);
/* extern void look_room_simple(dbref player, dbref room); */
extern void do_look_around(dbref player);
extern void do_look_at(dbref player, const char *name);
extern void do_examine(dbref player, const char *name);
extern void do_inventory(dbref player);
extern void do_find(dbref player, const char *name);
extern void do_owned(dbref player, const char *name);
extern void do_trace(dbref player, const char *name, int depth);
/*extern void exec_or_except(dbref player, dbref exception, const char
                           *msg, dbref thing);*/
extern void exec_or_notify(dbref player, dbref thing, const char *message);

/* From move.c */
extern void moveto(dbref what, dbref where);
extern void enter_room(dbref player, dbref loc, dbref exit);
extern void send_home(dbref thing);
extern int parent_loop_check(dbref source, dbref dest);
extern int can_move(dbref player, const char *direction);
extern void do_move(dbref player, const char *direction);
extern void do_get(dbref player, const char *what);
extern void do_drop(dbref player, const char *name);
#ifdef RECYCLE
extern void do_recycle(dbref player, const char *name);
#endif

/* From player.c */
extern dbref lookup_player(const char *name);
extern void do_password(dbref player, const char *old, const char *newobj);
extern void add_player(dbref who);
extern void delete_player(dbref who);

/* From predicates.c */
extern int can_link_to(dbref who, object_flag_type what_type, dbref where);
extern int can_link(dbref who, dbref what);
extern int could_doit(dbref player, dbref thing);
extern int can_doit(dbref player, dbref thing, const char *default_fail_msg);
extern int can_see(dbref player, dbref thing, int can_see_location);
extern int controls(dbref who, dbref what);
extern int restricted(dbref player, dbref thing, object_flag_type flag);
extern int payfor(dbref who, int cost);

/* From rob.c */
extern void do_kill(dbref player, const char *what, int cost);
extern void do_give(dbref player, const char *recipient, int amount);
extern void do_rob(dbref player, const char *what);

/* From set.c */
extern void do_name(dbref player, const char *name, char *newname);
extern void do_describe(dbref player, const char *name, const char *description);
extern void do_fail(dbref player, const char *name, const char *message);
extern void do_success(dbref player, const char *name, const char *message);
extern void do_drop_message(dbref player, const char *name, const char *message);
extern void do_osuccess(dbref player, const char *name, const char *message);
extern void do_ofail(dbref player, const char *name, const char *message);
extern void do_odrop(dbref player, const char *name, const char *message);
extern void do_lock(dbref player, const char *name, const char *keyname);
extern void do_unlock(dbref player, const char *name);
extern void do_unlink(dbref player, const char *name);
extern void do_chown(dbref player, const char *name, const char *newobj);
extern void do_set(dbref player, const char *name, const char *flag);

/* From speech.c */
extern void do_wall(dbref player, const char *message);
extern void do_gripe(dbref player, const char *message);
extern void do_say(dbref player, const char *message);
extern void do_page(dbref player, const char *arg1, const char *arg2);
extern void notify_except(dbref first, dbref exception, const char *msg);

/* From stringutil.c */
extern int string_compare(const char *s1, const char *s2);
extern int string_prefix(const char *string, const char *prefix);
extern const char *string_match(const char *src, const char *sub);
extern char *pronoun_substitute(dbref player, const char *str);

/* From utils.c */
extern int member(dbref thing, dbref list);
extern dbref remove_first(dbref first, dbref what);

/* From wiz.c */
extern void do_teleport(dbref player, const char *arg1, const char *arg2);
extern void do_force(dbref player, const char *what, char *command);
extern void do_stats(dbref player, const char *name);
extern void do_toad(dbref player, const char *name, const char *recip);
extern void do_boot(dbref player, const char *name);
#ifdef REGISTRATION
extern void do_pcreate(dbref player, const char *arg1, const char *arg2);
#endif /* REGISTRATION */

/* From boolexp.c */
extern int eval_boolexp(dbref player, struct boolexp *b, dbref thing);
extern struct boolexp *parse_boolexp(dbref player, const char *string);
extern struct boolexp *copy_bool(struct boolexp *old);

/* From unparse.c */
extern const char *unparse_object(dbref player, dbref object);
extern const char *unparse_boolexp(dbref player, struct boolexp *);

/* from property.c */
extern void add_property(dbref player, const char *type, const char *class, int value);
extern void remove_property(dbref player, const char *type);
extern int has_property(dbref player, const char *type, const char *class, int value);
extern int has_property_strict(dbref player, const char *type, const char *class, int value);
extern const char *get_property_class(dbref player, const char *type);
extern int get_property_value(dbref, const char *type);
extern int genderof(dbref player);
extern struct plist *copy_prop(dbref old);

/* From compress.c */
#ifdef COMPRESS
extern const char *compress(const char *);
extern const char *uncompress(const char *);
extern void init_compress(void);
#endif /* COMPRESS */

/* From edit.c */
extern void interactive(dbref player, const char *command);

/* From compile.c */
extern void do_compile(dbref in_player, dbref in_program);
extern void clear_primitives(void);
extern void init_primitives(void);

/* From interp.c */
extern int interp_loop(dbref player, dbref program, struct frame *fr);
extern int interp(dbref player, dbref program, dbref source);

/* declared in log.c */
extern void log_error();
extern void log_gripe();
extern void log_muf();
extern void log_conc();
extern void log_other();
