/* $Header: /belch_a/users/rearl/tinymuck/extract/RCS/merge.c,v 1.1 90/11/09 11:23:08 rearl Exp Locker: rearl $ */

#include "copyright.h"
#include "config.h"

#include "db.h"
#include "interface.h"

#define OK(x)	((x) >= 0)
#define ADJUST(x, y)	if (OK(x)) (x) += (y)
#define DOWNCASE(x) (lowercase[(x)])
extern const char *lowercase;
extern dbref db_size;

#include <stdio.h>
#include <ctype.h>

#include "params.h"

#define DB_MSGLEN 512

static const char *getstring_noalloc(FILE *f)
{
  static char buf[DB_MSGLEN];
  char *p;
  char c;
  
  fgets(buf, sizeof(buf), f);
  if (strlen(buf) == DB_MSGLEN - 1)
    {
      /* ignore whatever comes after */
      if (buf[DB_MSGLEN - 2] != '\n')
	while ((c = fgetc(f)) != '\n')
	  ;
    }
  for(p = buf; *p; p++) {
    if(*p == '\n') {
      *p = '\0';
      break;
    }
  }
  
  return buf;
}

static void db_grow(dbref newtop)
{
  struct object *newdb;
  
  if(newtop > db_top) {
    db_top = newtop;
    if(!db) {
      /* make the initial one */
      db_size = DB_INITIAL_SIZE;
      if((db = (struct object *)
	  malloc(db_size * sizeof(struct object))) == 0) {
	abort();
      }
    }
    
    /* maybe grow it */
    if(db_top > db_size) {
      /* make sure it's big enough */
      while(db_top > db_size) db_size *= 2;
      if((newdb = (struct object *)
	  realloc((void *) db,
		  db_size * sizeof(struct object))) == 0) {
	abort();
      } 
      db = newdb;
    }
  }
}

void clear_players(void)
{
}
void clear_primitives(void)
{
}
void init_primitives(void)
{
}
void add_player(dbref foo)
{
}
void do_compile(dbref foo, dbref bar)
{
}

void log_status(fmt, p1, p2, p3, p4)
     const char *fmt, *p1, *p2, *p3, *p4;
{
  fprintf(stderr, fmt, p1, p2, p3, p4);
}

struct macrotable
  *new_macro(const char *name, const char *definition, dbref player)
{
  struct macrotable *newmacro =
    (struct macrotable *)malloc (sizeof (struct macrotable));
  char buf[BUFFER_LEN];
  int i;
  
  for (i = 0; name[i]; i++)
    buf[i] = DOWNCASE(name[i]);
  buf[i] = '\0';
  newmacro -> name = alloc_string(buf);
  newmacro -> definition = alloc_string(definition);
  newmacro -> implementor = player;
  newmacro -> left = NULL;
  newmacro -> right = NULL;
  return (newmacro);
}

void adjust_boolexp(struct boolexp *old, dbref offset)
{
  if (!old)	return;

  switch (old -> type) {
  case BOOLEXP_AND:
  case BOOLEXP_OR:
    adjust_boolexp(old -> sub1, offset);
    adjust_boolexp(old -> sub2, offset);
    break;
  case BOOLEXP_NOT:
    adjust_boolexp(old -> sub1, offset);
    break;
  case BOOLEXP_CONST:
    ADJUST(old->thing, offset);
    break;
  }
}
    
dbref db_merge(FILE *f)
{
  dbref i;
  dbref oldtop;
  dbref newobj;
  dbref stuff;
  struct object *o;
  char buf[BUFFER_LEN];
  int  newformat;
  int	j;
  char  c;
  
  newformat = 0;
  oldtop = db_top;
  
  if ( (c = getc(f)) == '*') {
    (void) fgets(buf, sizeof (buf), f);
    if(!strcmp(buf, "**TinyMUCK DUMP Format***\n")) {
      newformat = 1;
    }
    else if (!strcmp(buf, "**Lachesis TinyMUCK DUMP Format***\n"))
      {
	newformat = 2;
      }
      else if (!strcmp(buf, "**Islandia TinyMUD DUMP Format***\n"))
	{
	  newformat = 3;
        }
    c = getc(f); /* get next char */
    printf("File format: %d\n",newformat);
  }
  
  for (i = 0;; i++) {
    switch (c) {
    case '#':
      /* another entry, yawn */
      if(i != getref(f)) {
	/* we blew it */
	return -1;
      }

      newobj = i + oldtop;
      db_grow(newobj + 1);

      /* read it in */
      o = DBFETCH(newobj);
      switch (newformat)
	{
	case 0:
	  db_read_object_old(f, o, newobj);
	  break;
	case 1:
	  db_read_object_new(f, o, newobj);
	  break;
	case 2:
	  db_read_object_lachesis(f, o, newobj);
	  break;
        case 3:
	  db_read_object_old(f, o, newobj);
	  getstring_noalloc(f);
	  getstring_noalloc(f);
	  getstring_noalloc(f);
	  break;
	}
      /* Fix all refs to old db numbers */
      ADJUST(o->location, oldtop);
      ADJUST(o->contents, oldtop);
      ADJUST(o->next, oldtop);
      adjust_boolexp(o->key, oldtop);

      switch (Typeof(newobj)) {
      case TYPE_THING:
	ADJUST(o->sp.thing.home, oldtop);
	ADJUST(o->sp.thing.actions, oldtop);
	ADJUST(OWNER(newobj), oldtop);
	break;

      case TYPE_ROOM:
	ADJUST(o->sp.room.dropto, oldtop);
	ADJUST(o->sp.room.exits, oldtop);
	ADJUST(OWNER(newobj), oldtop);
	break;

      case TYPE_EXIT:
	for (j = 0; j < o->sp.exit.ndest; j++) {
	  ADJUST((o->sp.exit.dest)[j], oldtop);
	}
	ADJUST(OWNER(newobj), oldtop);
	break;

      case TYPE_PLAYER:
	ADJUST(o->sp.player.home, oldtop);
	ADJUST(o->sp.player.actions, oldtop);
	for (stuff = 0; stuff < newobj; stuff++) {
	  if (Typeof(stuff) == TYPE_PLAYER && NAME(stuff)
	      && !string_compare(NAME(stuff), NAME(newobj))) {
	    fprintf(stderr,
		    "Player name conflict encountered with %s: %d and %d\n",
		    NAME(stuff), stuff, newobj);
	  }
	}
	ADJUST(OWNER(newobj), oldtop);
	break;

      case TYPE_PROGRAM:
      case TYPE_GARBAGE:
	ADJUST(OWNER(newobj), oldtop);
	break;
      }
      break;
    case '*':
      (void) fgets(buf, sizeof (buf), f);
      if(strcmp(buf, "**END OF DUMP***\n")) {
	return -1;
      } else {
	return db_top;
      }
    default:
      return -1;
      /* break; */
    }
    c = getc(f);
  } /* for */
}

int main(int argc, char **argv)
{
  FILE	*f;
  int	i;
  char old[BUFFER_LEN];
  char new[BUFFER_LEN];
  dbref oldtop;
  dbref objects;
  
  if ((f = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "Couldn't open %s!\n", argv[1]);
    exit (1);
  }
  printf("LOADING: %s\n", argv[1]);
  if (db_read(f) < 0) {
    fprintf(stderr, "Couldn't load %s!\n", argv[1]);
    exit (3);
  }
  fclose(f);
  printf("LOADING: %s (done) read %d objects\n", argv[1], db_top);
  for (i = 2; i < argc - 1; i++) {
    if ((f = fopen(argv[i], "r")) == NULL) {
      fprintf(stderr, "Couldn't open %s!\n", argv[i]);
      exit (i + 2);
    }
    printf("LOADING: %s\n", argv[i]);
    oldtop = db_top;
    if ((objects = db_merge(f)) < 0) {
      fprintf(stderr, "Couldn't load %s!\n", argv[i]);
      exit (i + 3);
    }
    printf("LOADING: %s (done) read %d objects\n", argv[i], objects - oldtop);
    fclose(f);
    for (objects = oldtop; objects < db_top; objects++) {
      if (Typeof(objects) == TYPE_PROGRAM) {
	sprintf(old, "muf.%s/%d.m", argv[i], objects - oldtop);
	sprintf(new, "muf/%d.m", objects);
	if (rename(old, new) < 0) {
	  sprintf(new, "rename of %s", old);
	  perror(new);
	}
      }
    }
  }
  if ((f = fopen(argv[i], "w")) == NULL) {
    fprintf(stderr, "Couldn't open %s!\n", argv[i]);
    exit (i + 4);
  }
  printf("DUMPING: %s\n", argv[i]);
  (void) db_write(f);
  printf("DUMPING: %s (done) wrote %d objects\n", argv[i], db_top);
  fclose(f);
  exit (0);
  return 1;
}
